# Copyright 2019 RnD Center "ELVEES", JSC

from lxml import etree  # type: ignore
import typing

CVATImageAnnotation = typing.Dict[str, typing.Any]


class BaseAnnotations:
    meta_type = ''

    def __init__(self):
        self.data = []
        self.alg_info = {}
        self.elem_types = ['box', 'points', 'polygon', 'polyline']
        self.box_points = ['xbr', 'xtl', 'ybr', 'ytl']
        self.meta_int_fields = ['id', 'height', 'width']
        self.meta_str_fields = ['name', 'label']
        self.elem_int_fields = ['z_order', 'frame']
        self.elem_bool_fields = ['occluded', 'outside', 'keyframe']

    def convert_types(self):
        for meta in self.data:
            for int_field in self.meta_int_fields:
                try:
                    meta[int_field] = int(meta[int_field])
                except KeyError:
                    pass
            for elem_type in self.elem_types:
                try:
                    elem_list = meta[elem_type]
                except KeyError:
                    continue
                for elem in elem_list:
                    if elem_type == 'box':
                        for point in self.box_points:
                            elem[point] = float(elem[point])
                    else:
                        points = elem['points']
                        for x, y in (coord.split(',') for coord in points.split(';')):
                            try:
                                elem['points'].append([float(x), float(y)])
                            except AttributeError:
                                elem['points'] = [[float(x), float(y)]]
                    for int_field in self.elem_int_fields:
                        try:
                            elem[int_field] = int(elem[int_field])
                        except KeyError:
                            pass
                    for bool_field in self.elem_bool_fields:
                        try:
                            elem[bool_field] = elem[bool_field] != '0'
                        except KeyError:
                            pass

    def load(self, file_):
        tree = etree.parse(file_)
        xml_annotations = tree.getroot()
        for xml_meta in xml_annotations.findall(self.meta_type):
            meta = dict(xml_meta.items())
            for xml_elem in xml_meta.getchildren():
                elem = dict(xml_elem.items())
                attr = {}
                for xml_attr in xml_elem.getchildren():
                    attr.update({xml_attr.values()[0]: xml_attr.text})
                elem.update({'attr': attr})
                try:
                    meta[xml_elem.tag].append(elem)
                except KeyError:
                    meta.update({xml_elem.tag: [elem]})
            self.data.append(meta)
        self.convert_types()
        try:
            for info_attr in xml_annotations.find("alg-info").getchildren():
                self.alg_info.update({info_attr.values()[0]: info_attr.text})
        except AttributeError:
            pass

    def dump(self, file_):
        try:
            self.data = sorted(self.data, key=lambda k: k['name'])
            for num, elem in enumerate(self.data):
                elem['id'] = num
        except KeyError:
            pass
        annotations = etree.Element('annotations')
        for meta in self.data:
            xml_meta = etree.SubElement(annotations, self.meta_type)
            for int_field in self.meta_int_fields:
                try:
                    xml_meta.set(int_field, str(meta[int_field]))
                except KeyError:
                    pass
            for str_field in self.meta_str_fields:
                try:
                    xml_meta.set(str_field, meta[str_field])
                except KeyError:
                    pass
            for elem_type in self.elem_types:
                try:
                    elem_list = meta[elem_type]
                except KeyError:
                    continue
                for elem in elem_list:
                    xml_elem = etree.SubElement(xml_meta, elem_type)
                    for bool_field in self.elem_bool_fields:
                        try:
                            xml_elem.set(bool_field, '1' if elem[bool_field] else '0')
                        except KeyError:
                            pass
                    for int_field in self.elem_int_fields:
                        try:
                            xml_elem.set(int_field, str(elem[int_field]))
                        except KeyError:
                            pass
                    try:
                        xml_elem.set('label', elem['label'])
                    except KeyError:
                        pass
                    if elem_type == 'box':
                        for point in self.box_points:
                            xml_elem.set(point, str(elem[point]))
                    else:
                        xml_elem.set('points',
                                     ';'.join(['{},{}'.format(x, y) for x, y in elem['points']]))
                    for key, value in elem['attr'].items():
                        xml_attr = etree.SubElement(xml_elem, 'attribute')
                        xml_attr.set('name', key)
                        xml_attr.text = value
        if self.alg_info:
            alg_info_attr = etree.SubElement(annotations, 'alg-info')
            for key, value in self.alg_info.items():
                info_attr = etree.SubElement(alg_info_attr, 'attribute')
                info_attr.set('name', key)
                info_attr.text = value
        tree = etree.ElementTree(annotations)
        file_.write((etree.tostring(tree, pretty_print=True)))


class ImageAnnotations(BaseAnnotations):
    """This is class, used to parse and serialize image annotations. For example,

    >>> import io
    >>> meta = ImageAnnotations()
    >>> xml = '''<annotations>
    ...   <image id="0" name="some_filename">
    ...     <box label="face" xtl="10" ytl="20" xbr="15" ybr="30" occluded="0">
    ...       <attribute name="gender">male</attribute>
    ...       <attribute name="face_id">some_id</attribute>
    ...     </box>
    ...     <polygon label="some_label" points="1,2;3,4;5,6" occluded="0">
    ...       <attribute name="gender">male</attribute>
    ...       <attribute name="face_id">some_id</attribute>
    ...     </polygon>
    ...   </image>
    ... </annotations>
    ... '''
    >>> meta.load(io.StringIO(xml))
    >>> meta.data[0]['name']
    'some_filename'
    >>> meta.data[0]['box'][0]['label']
    'face'
    >>> meta.data[0]['polygon'][0]['attr']['gender'] = 'female'
    """

    meta_type = 'image'


class TrackAnnotations(BaseAnnotations):
    """This is class, used to parse and serialize track annotations. For example,

    >>> import io
    >>> meta = TrackAnnotations()
    >>> xml = '''<annotations>
    ...   <track id="0" name="some_filename">
    ...     <box label="face" xtl="1" ytl="2" xbr="3" ybr="4" occluded="0" keyframe="1" outside="0">
    ...       <attribute name="gender">male</attribute>
    ...       <attribute name="face_id">some_id</attribute>
    ...     </box>
    ...     <polygon label="some_label" points="1,2;3,4;5,6" occluded="0" keyframe="1" outside="0">
    ...       <attribute name="gender">male</attribute>
    ...       <attribute name="face_id">some_id</attribute>
    ...     </polygon>
    ...   </track>
    ... </annotations>
    ... '''
    >>> meta.load(io.StringIO(xml))
    >>> meta.data[0]['name']
    'some_filename'
    >>> meta.data[0]['box'][0]['label']
    'face'
    >>> meta.data[0]['polygon'][0]['attr']['gender'] = 'female'
    """

    meta_type = 'track'
